module ArrowParser(PArr, tokenP, stringP, emptyP, anyP, 
    parse, parseAll, parseLazy,
    skip, iskip, option, optionm, between, many, many1, skipMany, skipMany1, 
    sepBy, sepBy1, sepEndBy, sepEndBy1, thenSepBy, 
    sepMaybeEndBy, sepMaybeEndBy1,
    chain, choice, (<|>), thenMany,
    oneOf, noneOf, ignore,
    digit, upper, lower, letter, alphaNum, newline, space, spaces,
    currPos, getState, putState,
    modifyA, consA,
    Attr(..)) where

import Prelude hiding ((.),id)
import Control.Monad
import Control.Monad.Error
import Control.Monad.State
import Control.Category
import Control.Arrow
import qualified DualSet as Set
import DualSet(DualSet)
import Data.List(stripPrefix)

{- TODO:
 - szybkie wyliczanie nastepnika przy alternatywie (mapą)
 - implementacja zbiorów pamiętająca przedziały
 -}

data Attr t = Attr t | AttrPH

instance Eq (Attr t) where
    _ == _ = True

instance Ord (Attr t) where
    compare _ _ = EQ

instance Show t => Show (Attr t) where
    showsPrec n (Attr t) = showsPrec n t
    showsPrec _ (AttrPH) = shows "_"

{- Strzalka parsujaca.
 - Argumenty:
 - 1 - czy przyjmuje ciag pusty
 - 2 - zbior mozliwych pierwszych znakow
 - 3 - funkcja parsujaca 
 -}

data PSt s t = PSt { tokens :: [t], tcount :: !Int, userstate :: s } 
   | PErr String

failure :: String -> PMon s t a
failure s = put (PErr s) >> return undefined

type PMon s t = State (PSt s t)

data (Show t, Ord t) => PArr s t a b = Parser Bool (DualSet t) (a -> PMon s t b)

instance (Show t, Ord t) => Category (PArr s t) where
    id = Parser True Set.empty return
    ~(Parser _ _ f2) . (Parser False s f1) = 
        Parser False s (f1 >=> f2)
    ~(Parser e s2 f2) . (Parser True s1 f1) =
        Parser e (Set.union s1 s2) (f1 >=> f2)

instance (Show t, Ord t) => Arrow (PArr s t) where
    arr f = Parser True Set.empty (return . f)
    first (Parser e s f) =
        Parser e s (runKleisli $ first $ Kleisli f)

instance (Show t, Ord t) => ArrowZero (PArr s t) where
    zeroArrow = Parser False Set.empty (\_ -> failure "zeroArrow")

instance (Show t, Ord t) => ArrowPlus (PArr s t) where
    (Parser e1 s1 f1) <+> (Parser e2 s2 f2) =
        Parser (e1 || e2) (Set.union s1 s2) f where
        f a = look >>= g a
        g a [] | e1 = f1 a
               | e2 = f2 a
               | otherwise = failure "unexpected EOF"
        g a (h:_) | h `Set.member` s1 = f1 a
                  | h `Set.member` s2 = f2 a
                  | e1 = f1 a
                  | e2 = f2 a
                  | otherwise = failure ("expected " ++ show (Set.union s1 s2) ++ ", found " ++ show h)

look :: PMon s t [t]
look = get >>= (return.tokens)

currPos :: (Ord t, Show t) => PArr s t () Int
currPos = Parser True Set.empty (\() -> get >>= (return.tcount))

getState :: (Ord t, Show t) => PArr s t () s
getState = Parser True Set.empty (\() -> get >>= (return.userstate))

putState :: (Ord t, Show t) => PArr s t s ()
putState = Parser True Set.empty (\s -> get >>= (\x -> put (x { userstate = s})))

setP :: (Ord t, Show t) => DualSet t -> PArr s t () t
setP s = Parser False s (\() -> get >>= f) where
    f (PSt (h:t) p st) | h `Set.member` s = put (PSt t (p+1) st) >> return h
                       | otherwise = failure ("expected " ++ show s ++ ", found " ++ show h)
    f _ = failure "unexpected EOF"

tokenP :: (Ord t, Show t) => t -> PArr s t () t
tokenP c = setP (Set.singleton c)

anyP :: (Ord t, Show t) => PArr s t () t
anyP = setP (Set.full)

oneOf :: (Ord t, Show t) => [t] -> PArr s t () t
oneOf l = setP (Set.fromList l)

noneOf :: (Ord t, Show t) => [t] -> PArr s t () t
noneOf l = setP (Set.complement $ Set.fromList l)

stringP :: (Ord t, Show t) => [t] -> PArr s t () [t]
stringP [] = arr (const [])
stringP s@(sh:st) = Parser False (Set.singleton sh) (\() -> get >>= f) where
    f (PSt l p st) = case stripPrefix s l of
            Just l' -> put (PSt l' (p+length s) st) >> return s

emptyP :: (Ord t, Show t) => PArr s t a a
emptyP = returnA

ignore :: (Ord t, Show t) => PArr s t a ()
ignore = arr (const ())

zeroP :: (Ord t, Show t) => PArr s t a b
zeroP = zeroArrow

parseLazy :: (Ord t, Show t) => PArr s t a b -> (a,[t],s) -> b
parseLazy (Parser _ _ f) (a,b,s) = fst $ runState (f a) (PSt b 0 s) 

parse :: (Ord t, Show t) => PArr s t a b -> (a,[t],s) -> Either String (b,[t],s)
parse (Parser _ _ f) (a,b,s) = case runState (f a) (PSt b 0 s) of
    (e, PErr f) -> Left f
    (a',PSt b' _ s') -> Right (a',b',s')

parseAll :: (Ord t, Show t) => PArr s t a b -> (a,[t],s) -> Either String (b,s)
parseAll p x = do
    (r,l,s) <- parse p x
    if l == [] then return (r,s) else throwError "failed to parse entire input"

skip :: (Ord t, Show t) => PArr s t a b -> PArr s t a a
skip p = emptyP &&& p >>^ fst

iskip :: (Ord t, Show t) => PArr s t () b -> PArr s t a a
iskip p = skip (ignore >>> p)

between :: (Ord t, Show t) => PArr s t () x -> PArr s t () y -> 
    PArr s t a b -> PArr s t a b
between l r c = iskip l >>> c >>> iskip r

option :: (Ord t, Show t) => b -> PArr s t a b -> PArr s t a b
option v p = p <+> arr (const v)

optionm :: (Ord t, Show t) => PArr s t a b -> PArr s t a (Maybe b)
optionm p = option Nothing (p >>^ Just)

many :: (Ord t, Show t) => PArr s t a b -> PArr s t a [b]
many p = sepBy p emptyP

many1 :: (Ord t, Show t) => PArr s t a b -> PArr s t a [b]
many1 p = sepBy1 p emptyP

thenMany :: (Ord t, Show t) => PArr s t a b -> PArr s t a b -> PArr s t a [b]
thenMany f p = thenSepBy f p emptyP

chain :: (Ord t, Show t) => PArr s t a a -> PArr s t a a
chain p = p >>> chain p <|> emptyP

skipMany :: (Ord t, Show t) => PArr s t a b -> PArr s t a a
skipMany = skip . many

skipMany1 :: (Ord t, Show t) => PArr s t a b -> PArr s t a a
skipMany1 = skip . many1

sepBy :: (Ord t, Show t) => PArr s t a b -> PArr s t a a -> PArr s t a [b]
sepBy p s = option [] (sepBy1 p s)
                
sepBy1 :: (Ord t, Show t) => PArr s t a b -> PArr s t a a -> PArr s t a [b]
sepBy1 p s = thenSepBy p p s

thenSepBy :: (Ord t, Show t) => PArr s t a b -> PArr s t a b -> 
    PArr s t a a -> PArr s t a [b]
thenSepBy f p s = f &&& option [] lol >>^ uncurry (:)
    where lol = (s >>> p) &&& option [] lol >>^ uncurry (:)

thenSepMaybeEndBy :: (Ord t, Show t) => PArr s t a b -> PArr s t a b -> 
    PArr s t a a -> PArr s t a [b]
thenSepMaybeEndBy f p s = f &&& option [] (s >>> option [] lol) >>^ uncurry (:)
    where lol = p &&& option [] (s >>> option [] lol) >>^ uncurry (:)

sepEndBy :: (Ord t, Show t) => PArr s t a b -> PArr s t b b -> PArr s t a [b]
sepEndBy p s = many (p >>> s)

sepEndBy1 :: (Ord t, Show t) => PArr s t a b -> PArr s t b b -> PArr s t a [b]
sepEndBy1 p s = many1 (p >>> s)

sepMaybeEndBy p s = option [] (sepMaybeEndBy1 p s)

sepMaybeEndBy1 p s = thenSepMaybeEndBy p p s

(<|>) :: (Ord t, Show t) => PArr s t a b -> PArr s t a b -> PArr s t a b
(<|>) = (<+>)
infixl 0 <|>

choice :: (Ord t, Show t) => [PArr s t a b] -> PArr s t a b
choice ps = foldl (<+>) zeroP ps

digit = oneOf ['0'..'9']
upper = oneOf ['A'..'Z']
lower = oneOf ['a'..'z']
letter = upper <+> lower
alphaNum = letter <+> digit
newline = tokenP '\n'
tab = tokenP '\t'
space = oneOf " \v\f\t\r\n"
spaces = skipMany space

modifyA f p = ((ignore >>> p) &&& returnA) >>^ uncurry f

consA :: (Show t, Ord t) => PArr s t () a -> PArr s t [a] [a]
consA = modifyA (:)

